#define _f_nums 3
/* constants above must be modified by the likelihood functions to fit */
#define MAXDATA 200
/* constants above must be modified by the number of observations */
#define REALZERO 0.0E+0
#define REALONE 1.0E+0
#define _ds_lambda_init 10.0
#define _ds_compress 0.5
#define _ds_reverse -1.0
#define _ds_expand 2.0
#define _ds_th 1.0E-15
#define _ds_itermax 10000

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

typedef struct vec {double p[_f_nums+1];} pvec; /* pvec.p[_f_nums] is return value of function */
static pvec _f_parms={0.03, 0.05, 0.01, 0.0}; /* plausible initial parameters required */

static long age[MAXDATA], n_data;
static double life[MAXDATA];

static pvec ivec[_f_nums+1]; /* unit vectors used in downhill simplex */
static char filename[80];

double f(pvec PP);
void sort(pvec *X);
long downhills(void);
long read_data(FILE *fp);

int main(int argc, char *argv[]) {

	FILE *ff;
	long i, j, k;
	int fnp;
	char *fx;
	
	if (argc<=1) {
		fputs("A data file name must be given.\n",stderr);
		fputs("Usage: clf datafile.dat\n\n The result is obtained in stdout\n",stderr);
		
		return 0;
	}
	else {
		strcpy(filename,argv[1]);
		ff=fopen(filename,"rt");
		n_data=read_data(ff);
		fclose(ff);
		
		if (argc>=3)
			_f_parms.p[0]=atof(argv[2]);
		if (argc>=4)
			_f_parms.p[1]=atof(argv[3]);
		if (argc>=5)
			_f_parms.p[2]=atof(argv[4]);
	}
	
	k=downhills();
	
	if (k==_ds_itermax)
		fprintf(stdout,"not convergent within %ld iterations\n",k);
	else
		fprintf(stdout,"a= %lf, b= %lf, c= %lf, RMSE=%e\n",ivec[0].p[0], ivec[0].p[1], ivec[0].p[2], ivec[0].p[3]);
	
	return 1;
}

long read_data(FILE *fp)
{
	long _rd_i, _rd_a, check;
	double _rd_w;

	_rd_i=0;
	do {
		check=fscanf(fp,"%d\t%lf\n",&_rd_a,&_rd_w);
		if (check!=EOF) {
			age[_rd_i]=_rd_a;
			life[_rd_i]=_rd_w;
			_rd_i++;
		}
	}
	while (check!=EOF);
	return _rd_i;
}

double f(pvec PP) {
	long ss;
	double aa, p1, p2, p3, pp, MSE;
	
	MSE=REALZERO;
	for (ss=0; ss<n_data; ss++) {
		aa = (double)age[ss];
		p1 = PP.p[0]*pow((aa/(105.0-aa)), 3.0);
		p2 = PP.p[1]*pow((exp(aa/(105.0-aa))-1.0), 0.5);
		p3 = PP.p[2]*(1.0-exp(-2.0*aa));
		pp = p1+p2+p3;
		pp = life[ss] - 1.0/(1.0+pp);
		MSE += pp*pp;
	}
	
	return sqrt(MSE/(double)n_data);
}

void sort(pvec *X) {
	pvec _s_swap;
	long _s_i, _s_k;
	
	for(_s_i=0; _s_i<_f_nums; _s_i++) {
		for(_s_k=_s_i; _s_k<=_f_nums; _s_k++) {
			if (X[_s_i].p[_f_nums]>X[_s_k].p[_f_nums]) {
				_s_swap=X[_s_i];
				X[_s_i]=X[_s_k];
				X[_s_k]=_s_swap;
			}
		}
	}
	return;
}

long downhills(void) {
	long _ds_i, _ds_j, _ds_iter=0;
	double _ds_rratio=1.0;
	pvec _ds_centroid, _ds_reflected, _ds_compressed;
	
	for (_ds_i=1; _ds_i<=_f_nums; _ds_i++) {
		for (_ds_j=0; _ds_j<_f_nums; _ds_j++) {
			ivec[0].p[_ds_j] = _f_parms.p[_ds_j];
			if (_ds_i==_ds_j+1)
				ivec[_ds_i].p[_ds_j] = ivec[0].p[_ds_j] + _ds_lambda_init;
			else
				ivec[_ds_i].p[_ds_j] = ivec[0].p[_ds_j];
		}
	} /* setting unit vectors */

	for(_ds_i=0; _ds_i<=_f_nums; _ds_i++) {
		ivec[_ds_i].p[_f_nums] = f(ivec[_ds_i]);
		#ifdef DDEBUG
		fprintf(stderr,"Initial L [%ld] th vector=%lf\n", _ds_i, ivec[_ds_i].p[_f_nums]);
		#endif
	}

	sort(ivec);

	while ((_ds_iter < _ds_itermax) && (_ds_rratio > _ds_th)) {
		if (ivec[0].p[_f_nums] != 0.0)
			_ds_rratio = 1.0 - ivec[0].p[_f_nums]/ivec[_f_nums].p[_f_nums];
/*			_ds_rratio = 2.0*(ivec[_f_nums].p[_f_nums]-ivec[0].p[_f_nums])/(fabs(ivec[_f_nums].p[_f_nums])+fabs(ivec[0].p[_f_nums])); */
		else
			_ds_rratio = 0.0;
		/* for the condition of exit loop */
		#ifdef DEBUG
			fprintf(stderr,"ratio = %lf in %ld th iteration\n", _ds_rratio, _ds_iter);
		#endif
		if (_ds_rratio <= _ds_th)
			break;
		/* calculating centroid of the vectors except the worst point */
		for (_ds_i=0; _ds_i<_f_nums; _ds_i++) {
			_ds_centroid.p[_ds_i] = REALZERO;
			for (_ds_j=0; _ds_j<_f_nums; _ds_j++) {
				_ds_centroid.p[_ds_i] += ivec[_ds_j].p[_ds_i]/(double)_f_nums;
			}
			#ifdef XXDEBUG
				fprintf(stderr,"centroid [%ld] = %lf in iter. %ld\n", _ds_i, _ds_centroid.p[_ds_i], _ds_iter);
			#endif
		}
		_ds_centroid.p[_f_nums] = f(_ds_centroid);
		#ifdef XDEBUG
			fprintf(stderr, "this centroid's f() = %lf\n", _ds_centroid.p[_f_nums]);
		#endif
		
		for (_ds_i=0; _ds_i<_f_nums; _ds_i++) {
			_ds_reflected.p[_ds_i] = 2.0*_ds_centroid.p[_ds_i] - ivec[_f_nums].p[_ds_i];
			#ifdef XXDEBUG
				fprintf(stderr,"reflected [%ld] = %lf in iter. %ld\n", _ds_i, _ds_reflected.p[_ds_i], _ds_iter);
			#endif
		}
		
		_ds_reflected.p[_f_nums] = f(_ds_reflected);
		#ifdef XDEBUG
			fprintf(stderr,"reflected f() = %lf in iter. %ld\n", _ds_reflected.p[_f_nums], _ds_iter);
		#endif
		
		if (fabs(_ds_reflected.p[_f_nums]-_ds_centroid.p[_f_nums])<_ds_th) _ds_rratio = 0.0;

		if (_ds_reflected.p[_f_nums] < ivec[_f_nums].p[_f_nums]) {
			for (_ds_i=0; _ds_i<=_f_nums; _ds_i++)
				ivec[_f_nums].p[_ds_i] = _ds_reflected.p[_ds_i];
		}
		
		if (ivec[_f_nums].p[_f_nums] <= ivec[0].p[_f_nums]) {
			/* when the reflected point becomes the best, the simplex must be expanded */
			for (_ds_i=0; _ds_i<_f_nums; _ds_i++)
				ivec[_f_nums].p[_ds_i] = 2.0*ivec[_f_nums].p[_ds_i] - _ds_centroid.p[_ds_i];
			ivec[_f_nums].p[_f_nums] = f(ivec[_f_nums]);
			#ifdef DDEBUG
			fprintf(stderr,"In iter. %ld, L [%ld] =%lf\n", _ds_iter, _f_nums, ivec[_f_nums].p[_f_nums]);
			#endif
		}
		else {
			if (ivec[_f_nums].p[_f_nums] > ivec[_f_nums-1].p[_f_nums]) {
				/* when the reflected point is still the worst, the simplex must be compressed */
				for (_ds_i=0; _ds_i<_f_nums; _ds_i++)
					_ds_compressed.p[_ds_i] = (ivec[_f_nums].p[_ds_i]+_ds_centroid.p[_ds_i])/2.0;
				/*	_ds_compressed.p[_ds_i] = (ivec[_f_nums].p[_ds_i]+ivec[0].p[_ds_i])/2.0; */
				_ds_compressed.p[_f_nums] = f(_ds_compressed);
				
				if (_ds_compressed.p[_f_nums] < ivec[_f_nums].p[_f_nums]) {
					for (_ds_i=0; _ds_i<=_f_nums; _ds_i++)
						ivec[_f_nums].p[_ds_i] = _ds_compressed.p[_ds_i];
				}
				
				if (ivec[_f_nums].p[_f_nums] > ivec[_f_nums-1].p[_f_nums]) { 
					/* when still worst, whole simplex must be compressed to the best point */
					for (_ds_i=1; _ds_i<=_f_nums; _ds_i++)
						for (_ds_j=0; _ds_j<_f_nums; _ds_j++)
							ivec[_ds_i].p[_ds_j] = (ivec[0].p[_ds_j] + ivec[_ds_i].p[_ds_j])/2.0;
					for(_ds_i=0; _ds_i<=_f_nums; _ds_i++)
						ivec[_ds_i].p[_f_nums] = f(ivec[_ds_i]);
				}
				else {
				/*	for (_ds_i=1; _ds_i<=_f_nums; _ds_i++)
						for (_ds_j=0; _ds_j<_f_nums; _ds_j++)
							ivec[_ds_i].p[_ds_j] = (ivec[0].p[_ds_j] + ivec[_ds_i].p[_ds_j])/2.0;
					for(_ds_i=0; _ds_i<=_f_nums; _ds_i++)
						ivec[_ds_i].p[_f_nums] = f(ivec[_ds_i]); */
				}
				#ifdef XXDEBUG
					fprintf(stderr,"In iter. %ld, L =%lf\n", _ds_iter, ivec[0].p[_f_nums]);
				#endif
			}
			else {
			/*	for (_ds_i=1; _ds_i<=_f_nums; _ds_i++)
					for (_ds_j=0; _ds_j<_f_nums; _ds_j++)
						ivec[_ds_i].p[_ds_j] = (ivec[0].p[_ds_j] + ivec[_ds_i].p[_ds_j])/2.0;
				for(_ds_i=0; _ds_i<=_f_nums; _ds_i++)
					ivec[_ds_i].p[_f_nums] = f(ivec[_ds_i]); */
			}
			#ifdef XXDEBUG
				fprintf(stderr,"In iter. %ld, L =%lf\n", _ds_iter, ivec[0].p[_f_nums]);
			#endif
		}
		
		sort(ivec);
		
		_ds_iter++;
	}
	
	return _ds_iter;
}
